(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: ksi.ml,v 1.13 2002/08/12 12:42:12 malekith Exp $
 *
 * Module implementation: Ksi
 * Description: Ksi intermediate language output implementation. We do
 * pretty-printing here, as well as some optimizations (but not very 
 * excessive :)
 *)

type id = string

(* Ksi type system *)
type ty =
          Void
        | Bool
        | Int of int            (* int is size *)
        | UInt of int
        | Float of int
        | Ptr of ty             (* ptr to ... *)
        | FuncPtr of ty * (ty list)
                                (* return_type * parameters *)
        | Ellipsis              (* `...' to be used only in function parms *)
        | Struct of id
        | Enum of id
        | Union of id
;;

(* C language types *)

type c_type_type =
(* size_of(c_type(C_{float,{long_,}double})) is only valid size for floating
 * point type. *)
          C_float
        | C_double
        | C_long_double

(* C integer types (are they needed for anything at all?) *)
        | C_char
        | C_short
        | C_int
        | C_long
        | C_long_long
        | C_uchar
        | C_ushort
        | C_uint
        | C_ulong
        | C_ulong_long
(* C pointer type, Ptr(Void) actually *)
        | C_ptr

(* attributes that associated with functions, variables and types *)
type attr =
          Aligned of int
        | Packed
        | Local         (* this is the default anyway... *)
        | Public
        | Extern
        | SuppressDebug
        | ConstFunc
        | PureFunc
        | StaticCtor
        | StaticDtor
        | MallocLike
        | NoReturn
        | NoLimitStack
        | NoCheckMemoryUsage
        | Artificial
        | Weak
        | OneOnly
        | Comdat
        | Section of string
;;

type binop =
          Eq            (* == *)
        | Ne            (* != *)
        | Ge            (* >= *)
        | Gt            (* > *)
        | Le            (* <= *)
        | Lt            (* < *)
        | Plus          (* + *)
        | Minus         (* - *)
        | Mult          (* * *)
        | RDiv          (* float / float *)
        | ExactDiv      (* division to use when no modulo is expected, like
                         * division of pointer difference by object size 
                         * in C. *)
        | RoundDiv      (* round result *)
        | FloorDiv      (* round result down *)
        | CeilDiv       (* round result up *)
        | TruncDiv      (* truncate result *)
        (* corresponding modulus operations *)
        | RoundMod | FloorMod | CeilMod | TruncMod
        
        | BitAnd        (* & *)
        | BitIOr        (* | *)
        | BitAndTc      (* sth strange *)
        | BitXor        (* ^ *)
        | Max
        | Min
        | LRotate
        | LShift        (* << *)
        | RRotate
        | RShift        (* >> *)
        (* ++/--, first operand is l-value to modify,
         * second operand is how much to inc/decrement by. *)
        | PostInc | PreInc 
        | PostDec | PreDec
        | TruthAndIf    (* && *)
        | TruthOrIf     (* || *)
        | TruthAnd      (* similar to && but not lazy *)
        | TruthOr       (* similar to || but not lazy *)
        | TruthXor      (* same, but ^ is used instead of | or & *)
;;

type unop =
          Negate                (* -X *)
        | BitNot                (* ~X *)
        | TruthNot              (* !X *)
        | Abs
        | Addr                  (* &X *)
        | Ffs                   (* position of first bit set *)
        | NonLvalue
;;

type exp =
    (* (const ...) term *)
      IntConst of ty * string
      (* string is decimal representation used to avoid problems with
       * different size of OCaml's int, int32 or int64 and target integer
       * size. It is Ksi responsibility to convert this correctly. *)
    | FloatConst of ty * string
    | BoolConst of bool                     (* There are no such problems with
                                             * bool... :^) *)
    | StringConst of string                 (* Ditto. *)
    | Null of ty                            (* null constant of type ty *)
    (* Single term of similar name: *)
    | Var of ty * id
    | StaticVar of ty * id
    | Bind of exp           (* create scope *)
    | Ref of id             (* reference variable by name *)
    | FuncRef of id         (* get address of named function *)
    | Modify of exp * exp   (* exp = exp *)
    | Component of exp * id (* exp.string *)
    | Arrow of exp * id     (* exp->string, shortcut *)
    (* this is shortcut to (return (modify (result) exp))
     * exp can be Nop, (return (nil)) is generated then *)
    | Return of exp
    | SizeOf of ty
    | Compound of exp * exp         (* exp, exp *)
    | CompoundList of exp list      (* exp1, exp2, ..., expn *)
    | Label of id
    | Goto of id
    | Loop of exp                   (* while (1) exp; *)
    | ExitLoop of (* bool *) exp    (* if (exp) break; *)
    | Nop
    | Indirect of exp               (* *exp *)
    | Cond of (* bool *) exp * exp * exp    (* exp ? exp : exp *)
    | Call of id * exp list             (* call named function, passing
                                         * parameters *)
    | FCall of id * exp list            (* call function that is
                                         * guaranteed not to have 
                                         * any side-effects *)
    | PtrCall of exp * exp list
    | PtrFCall of exp * exp list
    | BuiltIn of id * exp list
    | FBuiltIn of id * exp list
    | Convert of ty * exp
    | Block of id * exp                 (* create named block *)
    | Exit of id                        (* jump to end of named block *)
    | Binary of binop * exp * exp
    | Unary of unop * exp
    | TryFinally of exp * exp
    | Line of int * exp                     (* annotate expression with
                                             * line number information. *)
    | Comment of string * exp               (* annotate output Ksi code
                                             * with comment *)
;;

        
type decl =
          FuncProto of (attr list) * ty * id * (ty list)
        | FuncProtoNamed of (attr list) * ty * id * ((ty * id) list)
        | Func of (attr list) * ty * id * (ty * id) list * exp
        | TopComment of string          (* output comment *)
        | StructDefProto of id
        | StructDef of id * attr list * (ty * id) list
        | UnionDefProto of id
        | UnionDef of id * attr list * (ty * id) list
        | Global of attr list * ty * id
        | InputFile of string
;;


exception Conf_error of string
exception Sorry of string
;;

let make_id s = s;;

let debug s = print_string s; print_newline ();;

let output_header out_ch =
  output_string out_ch 
        ("// Ksi output generated by OCaml interface\n" ^
         "!version 1\n\n" ^
         "[\n")
;;

let output_trailer out_ch =
  output_string out_ch 
        ("]\n" ^
         "// EOF\n")
;;

(* information about underlying machine... *)

let c_float_bits = ref 0
let c_double_bits = ref 0
let c_long_double_bits = ref 0
let c_char_bits = ref 0
let c_short_bits = ref 0
let c_int_bits = ref 0
let c_long_bits = ref 0
let c_long_long_bits = ref 0
let c_uchar_bits = ref 0
let c_ushort_bits = ref 0
let c_uint_bits = ref 0
let c_ulong_bits = ref 0
let c_ulong_long_bits = ref 0
let c_ptr_bits = ref 0
;;

let bits_of_c_type t =
  match t with
      C_float -> c_float_bits
    | C_double -> c_double_bits
    | C_long_double -> c_long_double_bits
    | C_char -> c_char_bits
    | C_short -> c_short_bits
    | C_int -> c_int_bits
    | C_long -> c_long_bits
    | C_long_long -> c_long_long_bits
    | C_uchar -> c_uchar_bits
    | C_ushort -> c_ushort_bits
    | C_uint -> c_uint_bits
    | C_ulong -> c_ulong_bits
    | C_ulong_long -> c_ulong_long_bits
    | C_ptr -> c_ptr_bits
;;

let c_type t =
  let size = !(bits_of_c_type t) in
  match t with
      C_float -> Float(size)
    | C_double -> Float(size)
    | C_long_double -> Float(size)
    | C_char -> Int(size)
    | C_short -> Int(size)
    | C_int -> Int(size)
    | C_long -> Int(size)
    | C_long_long -> Int(size)
    | C_uchar -> UInt(size)
    | C_ushort -> UInt(size)
    | C_uint -> UInt(size)
    | C_ulong -> UInt(size)
    | C_ulong_long -> UInt(size)
    | C_ptr -> Ptr(Void)
;;

exception Bad_keyword
;;

let string_to_c_type s =
  match s with
      "c_float_bits" -> C_float
    | "c_double_bits" -> C_double
    | "c_long_double_bits" -> C_long_double
    | "c_char_bits" -> C_char
    | "c_short_bits" -> C_short
    | "c_int_bits" -> C_int
    | "c_long_bits" -> C_long
    | "c_long_long_bits" -> C_long_long
    | "c_uchar_bits" -> C_uchar
    | "c_ushort_bits" -> C_ushort
    | "c_uint_bits" -> C_uint
    | "c_ulong_bits" -> C_ulong
    | "c_ulong_long_bits" -> C_ulong_long
    | "c_ptr_bits" -> C_ptr
    | _ -> raise Bad_keyword
;;

let ref_target_bytes_bigendian = ref false
let ref_target_words_bigendian = ref false
let ref_ksi_version = ref 0
;;

let target_bytes_bigendian () = !ref_target_bytes_bigendian
let target_words_bigendian () = !ref_target_words_bigendian
let ksi_version () = !ref_ksi_version
;;

let read_target_conf prefix =
  let get_int st =    
    match Stream.next st with
        Genlex.Int(i) -> i
      | Genlex.Ident(s) -> raise(Conf_error ("expected int, got " ^ s))
      | _ -> raise(Conf_error "no int after keyword") in
   let read_conf st tok =
      match tok with
          Genlex.Ident("words_bigendian") -> 
            if get_int st == 0 then 
              ref_target_words_bigendian := false
            else
              ref_target_words_bigendian := true
        | Genlex.Ident("bytes_bigendian") -> 
            if get_int st == 0 then 
              ref_target_bytes_bigendian := false
            else
              ref_target_bytes_bigendian := true
        | Genlex.Ident("ksi_version") -> 
            ref_ksi_version := get_int st
        | Genlex.Ident(s) -> 
            (try 
               let r = (bits_of_c_type (string_to_c_type s)) in 
               r := get_int st
             with Bad_keyword -> let _ = Stream.next st in ())
        | _ -> raise(Conf_error "unknown token in input file") in
  let lex_st = ((Genlex.make_lexer []) 
                (Stream.of_channel (open_in (prefix ^ "/target_conf")))) in
  Stream.iter (read_conf lex_st) lex_st
;;

let ptroff () = Int(!c_ptr_bits)
let uptroff () = UInt(!c_ptr_bits)
;;

let size_of t =
  match t with
          Void -> 0
        | Bool -> !c_int_bits
        | Int(size) -> size
        | UInt(size) -> size
        | Float(size) -> size
        | Ptr(_) -> !c_ptr_bits
        | FuncPtr(_) -> !c_ptr_bits
        | Ellipsis -> assert false
        | Struct(_) -> raise(Sorry "size_of Struct")
        | Enum(_) -> raise(Sorry "size_of Struct")
        | Union(_) -> raise(Sorry "size_of Struct")
;;

(* and now, the funny thing :^) *)

type term = 
    T of string * term list             (* (string ...) *)
  | TList of term list                  (* [...] *)
  | TArg of string                      (* string (without `()') *)
  | TComment of string * term
  | TLine of int * term
;;

let one_int i = TArg(string_of_int(i))
;;

(* quote given string, output list of terms *)

let rec convert_string s =
  let chunk = 20 in
  if String.length s > chunk then
    (TArg("\"" ^ (String.escaped (String.sub s 0 chunk)) ^ "\".")) ::
    (convert_string (String.sub s chunk (String.length s - chunk)))
  else
    [TArg("\"" ^ (String.escaped s) ^ "\"")]
;;

let convert_attrs attrs = 
  let convert_attr a =
    match a with
        Aligned(i) -> T("aligned", [one_int i])
      | Packed -> T("packed", [])
      | Local -> T("local", [])
      | Public -> T("public", [])
      | Extern -> T("extern", [])
      | SuppressDebug -> T("suppress_debug", [])
      | ConstFunc -> T("const_func", [])
      | PureFunc -> T("pure_func", [])
      | StaticCtor -> T("static_ctor", [])
      | StaticDtor -> T("static_dtor", [])
      | MallocLike -> T("malloc_like", [])
      | NoReturn -> T("no_return", [])
      | NoLimitStack -> T("no_limit_stack", [])
      | NoCheckMemoryUsage -> T("no_check_memory_usage", [])
      | Artificial -> T("artificial", [])
      | Weak -> T("weak", [])
      | OneOnly -> T("one_only", [])
      | Comdat -> T("comdat", [])
      | Section(name) -> T("section", convert_string name)
in TList (List.map convert_attr attrs)
;;


let rec convert_unnamed_parms parms = 
  let rec cnt = ref 0
  and convert_unnamed_parm t =
    incr cnt; 
    match t with
    | Ellipsis -> T("ellipsis", [])
    | _ -> T("var", [convert_type t; TArg("_" ^ string_of_int(!cnt))])
  in TList (List.map convert_unnamed_parm parms)
and convert_type t =
  match t with
      Void -> T("void", [])
    | Bool -> T("bool", [])
    | Int(size) -> T("int", [one_int size])
    | UInt(size) -> T("uint", [one_int size])
    | Float(size) -> T("float", [one_int size])
    | Ptr(o) -> T("ptr", [convert_type o])
    | FuncPtr(ret, args) -> T("func_ptr", [convert_type ret; 
                                           TList (List.map convert_type
                                                           args)])
    | Ellipsis -> T("ellipsis", [])
    | Struct(name) -> T("struct", [TArg(name)])
    | Enum(name) -> T("enum", [TArg(name)])
    | Union(name) -> T("union", [TArg(name)])
;;

let convert_var (t, name) = T("var", [convert_type t; TArg(name)])
;;

let tnil = T("nil", [])
;;

let rec convert_exp exp =
  let bin_name b =
    match b with
          Eq -> "eq"
        | Ne -> "ne"
        | Ge -> "ge"
        | Gt -> "gt"
        | Le -> "le"
        | Lt -> "lt"
        | Plus -> "plus"
        | Minus -> "minus"
        | Mult -> "mult"
        | RDiv -> "rdiv"
        | ExactDiv -> "exact_div"
        | RoundDiv -> "round_div"
        | FloorDiv -> "floor_div"
        | CeilDiv -> "ceil_div"
        | TruncDiv -> "trunc_div"
        | TruncMod -> "trunc_mod"
        | RoundMod -> "round_mod"
        | FloorMod -> "floor_mod"
        | CeilMod -> "ceil_mod"
        | BitAnd -> "bit_and"
        | BitIOr -> "bit_ior"
        | BitAndTc -> "bit_andtc"
        | BitXor -> "bit_xor"
        | Max -> "max"
        | Min -> "min"
        | LRotate -> "lrotate"
        | LShift -> "lshift"
        | RRotate -> "rrotate"
        | RShift -> "rshift"
        | PostInc  -> "postinc"
        | PreInc  -> "preinc"
        | PostDec  -> "postdec"
        | PreDec -> "predec"
        | TruthAndIf -> "truth_andif"
        | TruthOrIf -> "truth_orif"
        | TruthAnd -> "truth_and"
        | TruthOr -> "truth_or"
        | TruthXor -> "truth_xor"
  and un_name u = 
    match u with
          Negate -> "negate"
        | BitNot -> "bit_not"
        | TruthNot -> "truth_not"
        | Abs -> "abs"
        | Addr -> "addr"
        | Ffs -> "ffs"
        | NonLvalue -> "non_lvalue"
  in
    match exp with
        IntConst(t, v) -> T("const", [convert_type t; TArg(v)])
      | FloatConst(t, v) -> 
          T("const", [convert_type t; TArg(v)])
      | StringConst(s) -> 
          T("const", (convert_type (Ptr(UInt(8)))) :: (convert_string s))
      | BoolConst true -> T("const", [convert_type Bool; one_int 1])
      | BoolConst false -> T("const", [convert_type Bool; one_int 0])
      | Null(t) -> T("const", [convert_type t; one_int 0])
      | Var(t, n) -> convert_var (t, n)
      | SizeOf(t) -> T("size_of", [convert_type t])
      | StaticVar(t, n) -> T("static_var", [convert_type t; TArg(n)])
      | Bind(e) -> T("bind", [convert_exp e])
      | Ref(n) -> T("ref", [TArg(n)])
      | FuncRef(n) -> T("func_ref", [TArg(n)])
      | Modify(e1, e2) -> T("modify", [convert_exp e1; convert_exp e2])
      (* this is shortcut to (return (modify (result) e))
       * exp can be Nop, (return (nop)) is generated then *)
      | Return(Nop) -> T("return", [T("nop", [])])
      | Return(e) -> 
          T("return", [T("modify", [T("result", []); convert_exp e])])
      | Compound(e1, e2) -> T("compound", [TList [convert_exp e1; 
                                                 convert_exp e2]])
      | CompoundList([]) -> T("nop", [])
      | CompoundList(lst) -> T("compound", [TList (List.map convert_exp lst)])
      | Label(n) -> T("label", [TArg(n)])
      | Goto(n) -> T("goto", [TArg(n)])
      | Loop(e) -> T("loop", [convert_exp e])
      | ExitLoop(e) -> T("exit_loop", [convert_exp e])
      | Nop -> T("nop", [])
      | Indirect(e) -> T("indirect", [convert_exp e])
      | Cond(e1, e2, e3) -> 
          T("cond", [convert_exp e1; convert_exp e2; convert_exp e3])
      | Call(name, args) ->
          T("call", [TArg(name); TList(List.map convert_exp args)])
      | FCall(name, args) ->
          T("fcall", [TArg(name); TList(List.map convert_exp args)])
      | PtrCall(e, args) ->
          T("ptr_call", [convert_exp e; 
                         TList(List.map convert_exp args)])
      | PtrFCall(e, args) ->
          T("ptr_fcall", [convert_exp e; 
                          TList(List.map convert_exp args)])
      | BuiltIn(name, args) ->
          T("builtin", [TArg(name); 
                        TList(List.map convert_exp args)])
      | FBuiltIn(name, args) ->
          T("fbuiltin", [TArg(name); 
                         TList(List.map convert_exp args)])
      | Convert(t, e) -> T("convert", [convert_type t; convert_exp e])
      | Block(n, e) -> T("block", [TArg(n); convert_exp e])
      | TryFinally (e1, e2) -> T("try_finally", [convert_exp e1; 
                                                 convert_exp e2])
      | Exit(n) -> T("exit", [TArg(n)])
      | Binary(op, e1, e2) -> 
          T(bin_name op, [convert_exp e1; convert_exp e2])
      | Unary(op, e) -> T(un_name op, [convert_exp e])
      | Line(i, e) -> TLine(i, convert_exp e)
      | Component(e, n) -> T("component", [convert_exp e; TArg(n)])
      | Arrow(e, n) -> T("component", [T("indirect", [convert_exp e]); TArg(n)])
      | Comment(s, e) -> TComment(s, convert_exp e)
;;

(* ok, now we'll setup our pretty-printer *)

let infty = 78;; (* for textual term representation length *)

exception Infty;;

let term_len t =
  let rec term_len_helper t =
    let rec list_len lst =
      match lst with
          [] -> 0
        | x::xs -> 
          let l = (term_len_helper x) + (list_len xs) + 1 in 
          if l >= infty then raise Infty else l in
    match t with
        T(name, lst) -> (list_len lst) + 2 + (String.length name) 
      | TList lst -> (list_len lst) + 2 
      | TArg(name) -> (String.length name)
      | TComment(_, _) -> raise Infty 
      | TLine(_, _) -> assert false in
  let len = (try term_len_helper t with Infty -> infty) in
    if len >= infty then
      infty
    else
      len
;;

let speedy_writer out_ch t =
  let rec f t = 
    match t with
    | T (name, lst) -> 
        output_string out_ch "(";
        output_string out_ch name;
        output_string out_ch "\n";
        List.iter f lst;
        output_string out_ch ")\n"
    | TList lst -> 
        output_string out_ch "[";
        List.iter f lst;
        output_string out_ch "]\n"
    | TArg name -> 
        output_string out_ch name;
        output_string out_ch " "
    | TComment (s, exp) ->
        output_string out_ch "// ";
        output_string out_ch s;
        output_string out_ch "\n";
        f exp
    | TLine(_, _) -> assert false in
  f t
;;

let fast_writer = ref false
;;

let rec writer t =
  match t with
    | T(name, []) -> "(" ^ name ^ ")"
    | T(name, lst) -> "(" ^ name ^ " " ^
                      (String.concat " " (List.map writer lst)) ^ ")"
    | TList lst -> "[" ^ (String.concat " " (List.map writer lst)) ^ "]"
    | TArg(name) -> name
    | _ -> assert false
and write_t_or_tlist out_ch ind apnd t =
  let rec map lst fnc apnd =
    match lst with
        [x] -> fnc apnd x
      | x::xs -> fnc "" x; map xs fnc apnd
      | [] -> assert false in
  let (char_open, char_close, name, lst) = 
    match t with
        T(n, l) -> ("(", ")", n, l)
      | TList l -> ("[", "]", "", l)
      | _ -> assert false
  and new_ind = if ind > (infty / 2) then ind else (ind + 1) in
  (output_string out_ch ((String.make ind ' ') ^ char_open ^ name ^ "\n");
   match lst with
      [] -> output_string out_ch (char_close ^ apnd ^ "\n")
    | _ -> map lst (line_writer out_ch new_ind) (char_close ^ apnd))
and line_writer out_ch ind apnd t =
  if (term_len t) >= (infty - ind) then
    match t with
        TComment(s, exp) ->
          (output_string out_ch ("// " ^ s ^ "\n");
           line_writer out_ch ind apnd exp)
      | T(_, _) -> write_t_or_tlist out_ch ind apnd t
      | TList _ -> write_t_or_tlist out_ch ind apnd t
      | TArg(s) -> 
          (output_string out_ch (String.make ind ' ');
           output_string out_ch s;
           output_string out_ch "\n")
      | TLine(_, _) -> assert false
  else
    output_string out_ch ((String.make ind ' ') ^ (writer t) ^ apnd ^ "\n")
and write_term out_ch t = 
  if !fast_writer then
    speedy_writer out_ch t
  else
    line_writer out_ch 0 "" t
;;

(* Remove redundant TLine terms *)
exception Cannot_simplify
let rec simplify t =
  let rec mapper ln r tt =
    match tt with
        TLine(x, e)::ts ->
          if ln != -1 && ln != x then
            raise Cannot_simplify
          else
            mapper x (e::r) ts
      | t::ts -> mapper ln (t::r) ts
      | [] -> (List.rev r, ln) in
  match t with
        TComment(s, exp) ->
          let sexp = simplify exp in
            (match sexp with
                 TLine(l, e) -> TLine(l, TComment(s, e))
               | _ -> TComment(s, sexp))
      | T(s, a) ->
          let a' = List.map simplify a in
          (try
              let (a'', ln) = mapper (-1) [] a' in
              if ln != -1 then
                TLine(ln, T(s, a''))
              else
                T(s, a'')
           with Cannot_simplify -> T(s, a'))
      | TList a ->
          let a' = List.map simplify a in
          (try
              let (a'', ln) = mapper (-1) [] a' in
              if ln != -1 then
                TLine(ln, TList a'')
              else
                TList a''
           with Cannot_simplify -> TList a')
      | TArg(s) -> TArg(s)
      | TLine(ln, e) ->
           let e' = simplify e in
           match e' with
               TLine(ln', e'') when (ln == ln') -> TLine(ln, e'')
             | _ -> TLine(ln, e')
let rec convert_tlines t =
  match t with
      TComment(s, exp) -> TComment(s, convert_tlines exp)
    | T(s, a) -> T(s, List.map convert_tlines a)
    | TList a -> TList (List.map convert_tlines a)
    | TArg(s) -> TArg(s)
    | TLine(ln, e) -> T("wfl", [one_int ln; convert_tlines e])
;;

let output_decl out_ch d =
  let convert_decl d =
    match d with
          FuncProto(attrs, ret_type, name, params) -> 
            T("func", [convert_attrs attrs;
                       convert_type ret_type;
                       TArg(name);
                       convert_unnamed_parms params;
                       tnil])
        | FuncProtoNamed(attrs, ret_type, name, params) -> 
            T("func", [convert_attrs attrs;
                       convert_type ret_type;
                       TArg(name);
                       TList (List.map convert_var params);
                       tnil])
        | Func(attrs, ret_type, name, parms, exp) -> 
            let exp' = convert_tlines (simplify (convert_exp exp)) in
            let exp'' = 
              match exp' with
                  T("bind", _) -> exp'
                | e -> T("bind", [e]) in
            T("func", [convert_attrs attrs; 
                       convert_type ret_type; 
                       TArg(name);
                       TList (List.map convert_var parms);
                       exp''])
        | StructDefProto(name) -> T("structdef", [TArg(name); tnil])
        | UnionDefProto(name) -> T("uniondef", [TArg(name); tnil])
        | StructDef(name, attrs, fields) ->
            T("structdef", [convert_attrs attrs; 
                            TArg (name); 
                            TList (List.map convert_var fields)])
        | UnionDef(name, attrs, fields) ->
            T("uniondef", [convert_attrs attrs; 
                           TArg (name);
                           TList (List.map convert_var fields)])
        | Global(attrs, t, n) ->
            T("global", [convert_attrs attrs; 
                         T("var", [convert_type t; TArg(n)])])
        | InputFile(name) ->
            T("input_file", convert_string name)
        | TopComment(_) -> assert false
in
  match d with
      TopComment(s) -> output_string out_ch ("// " ^ s ^ "\n")
    | _ -> 
      let converted = (convert_decl d) in
      ((* debug "got converted"; *)
       write_term out_ch converted)
;;

