(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: main.ml,v 1.12 2002/08/12 12:42:12 malekith Exp $
 *
 * Module implementation: Main
 * Description: Driver for Gont compiler.
 *)

let compile_only = ref false
let generate_ksi_only = ref false
let dump_flag = ref false
let ciface_flag = ref false
let save_temps = ref false
let ksi_compiler = ref "xgcc"
let ksi_opts : string list ref = ref []
let input_files : (bool * string) list ref = ref []
let verbose_flag = ref false
let w_error_flag = ref false
let gia_mode = ref false
let ga_mode = ref false
let output_file = ref "a.out"
let link_aux = ref "gont-link-aux.ksi"
let link_init = ref "gont_init"
let link_fini = ref "gont_fini"
let link_main = ref "main"
let just_make_aux = ref false
let ldopts = ref ""
let link_search_path : string list ref = ref []
let iface_search_path : string list ref = ref []
let prefix = ref Config.prefix
;;

let verb s = if !verbose_flag then (print_endline s; flush stdout) else ()
let err s = prerr_endline ((Array.get Sys.argv 0) ^ ": " ^ s); flush stderr
;;

let xf = Printf.sprintf

let rm x = try Sys.remove x with Sys_error _ -> ()
;;

let split_filename s =
  let (dir, file) = (Filename.dirname s, Filename.basename s) in
  let base = 
    try
      Filename.chop_extension file
    with Invalid_argument _ -> file in
  if base = file then begin
      err (xf "expected filename with extension, got `%s'" file);
      raise Exit
  end else
    let ext = String.sub file ((String.length base) + 1)
                              (String.length file - String.length base - 1) in
    (dir, base, ext)
;;

let execute argv =
  let cmd = String.concat " " argv in
  verb (xf "executing `%s'" cmd);
  let ret = Sys.command cmd in
  if ret != 0 then
    begin
      err (xf "`%s' failed, exit code = %d" cmd ret);
      raise Exit
    end
  else
    ()
;;

let compile_file gf = 
  let compile_ksi i o =
    verb (xf "compiling Ksi file `%s' -> `%s'" i o);
    execute ((!ksi_compiler :: !ksi_opts) @ [ "-o"; o; "-c"; i ]) in
  let compile_gont iface i o go m =
    let info = {Gont.ci_target_prefix = !prefix;
                Gont.ci_input_file = i;
                Gont.ci_output_file = o;
                Gont.ci_goutput_file = go;
                Gont.ci_module_name = m} in
    verb (xf "compiling %sGont file `%s' -> `%s'" 
             (if iface then "iface " else "") i o);
    let e = 
      if iface then 
        Gont.compile_iface info
      else
        Gont.compile info in
    let e =
      if e == 0 && !w_error_flag && !Gontutil.warnings_count != 0 then
        (err "warnings are being treated as errors";
         1)
      else
         e in
    if e != 0 then
      begin
        (if (not !save_temps) then rm o);
        exit 1
      end
    else 
       () in
  let (greedy, f) = gf in
  let (dir, basename, suffix) = split_filename f in
  let predir = Filename.concat dir in
  let ret = 
    match suffix with
        "g" -> 
          let aux = predir (basename ^ ".ksi") in
          let go = predir (basename ^ ".go") in
          let out = predir (basename ^ ".o") in
          let m = String.capitalize basename in
          compile_gont false f aux go m;
          if !generate_ksi_only then
            aux
          else
            begin
              begin
                try
                  compile_ksi aux out
                with x -> rm go; rm out; raise x
              end;
              if not !save_temps then rm aux;
              go
            end
      | "ga"
      | "go"
      | "a"
      | "o" -> f
      | "ksi" ->
          if !generate_ksi_only then
            f
          else
            let out = predir (basename ^ ".o") in
            (compile_ksi f out; out)
      | "gi" ->
          let out = predir (basename ^ ".gio") in
          let m = String.capitalize basename in
          compile_gont true f out "" m; out
      | _ ->
          err (xf "don't know what to do with `%s'" f);
          exit 1 in
   (greedy, ret)
;;

let link fs =
  let files = (false, "core.ga") :: (false, "std.ga") :: fs in
  let search_path = ("." :: !link_search_path) @ 
                    [!prefix ^ "/stdlib"; !prefix ^ "/corelib"] in
  let li = {Gontlink.li_target_prefix = !prefix;
            Gontlink.li_search_path = search_path;
            Gontlink.li_aux_ksi_file = !link_aux;
            Gontlink.li_init_name = !link_init;
            Gontlink.li_fini_name = !link_fini;
            Gontlink.li_main_name = !link_main;
            Gontlink.li_file_list = files;
            Gontlink.li_ldopts = !ldopts} in
  let o = !output_file in
  let nam (g, s) = if not g then s ^ ":lib" else s in
  verb (xf "linking `%s' -> `%s'"
           (String.concat "', `" (List.map nam files)) 
           o);
  let (ldopts, file_list) = 
    try 
      Gontlink.link li 
    with Gontutil.Error s -> 
      err ("link: " ^ s);
      if not !save_temps then rm !link_aux;
      exit 1 in
  let (_, obj) = compile_file (true, !link_aux) in
  try
    execute ((!ksi_compiler :: !ksi_opts) @ 
            [ "-o"; o; obj ] @ file_list @ [Config.libgc; ldopts]);
    (if not !save_temps then rm !link_aux);
    rm obj
  with Exit ->
    (if not !save_temps then rm !link_aux);
    rm obj;
    rm o;
    exit 1
;;

open Arg

let subst_plus s =
  if String.length s >= 1 && String.get s 0 = '+' then
    !prefix ^ "/" ^ (String.sub s 1 (String.length s - 1))
  else
    s

let main () =
  let add_ksi_opt x = 
     ksi_opts := x :: !ksi_opts in
  let add_greedy_input_file x =
     input_files := (true, x) :: !input_files in
  let add_ungreedy_input_file x =
     input_files := (false, x) :: !input_files in
  let prog = (Array.get Sys.argv 0) in
  let help_str = 
        prog ^ " - compile and link Gont files\n\n\
        If input file name matches *.g, it is assumed to be Gont source\n\
        file, is compiled to *.ksi file, which is in turn, passed to Ksi\n\
        compiler. *.ksi files are passed directly to Ksi compiler.\n" in
  let opt = [
    ("-c", Set compile_only, "           just compile");
    ("-save-temps", Set save_temps, "  save temporary files");
    ("-ksi", String add_ksi_opt, "OPT      pass option to ksi compiler");
    ("-ksicc", String (fun s -> ksi_compiler := s), "PROG   set Ksi compiler");
    ("-O0", Unit (fun () -> add_ksi_opt "-O0"), "          same as -ksi -O0");
    ("-O1", Unit (fun () -> add_ksi_opt "-O1"), "          same as -ksi -O1");
    ("-O2", Unit (fun () -> add_ksi_opt "-O2"), "          same as -ksi -O2");
    ("-O3", Unit (fun () -> add_ksi_opt "-O3"), "          same as -ksi -O3");
    ("-Werror", Set w_error_flag, "      treat warnings as fatal");
    ("-g", Unit (fun () -> add_ksi_opt "-g"), "           same as -ksi -g");
    ("-S", Set generate_ksi_only, "           just generate .ksi file \
                                   and stop");
    ("-o", String (fun s -> output_file := s), "OUTFILE    set linker output \
                                                file name");
    ("-B", String (fun s -> prefix := s ), "PREFIX     override prefix (\
                                            default " ^ !prefix ^ ")");
    ("-v", Set verbose_flag, "           be verbose");
    ("-dump", Set dump_flag, "        dump named modules");
    ("-ciface", Set ciface_flag, "      dump C interface of named modules");
    ("-gia", Set gia_mode, "         operate in make-iface-archive mode");
    ("-ga", Set ga_mode, "          operate in make-impl-archive mode");
    ("-ugly", Set Ksi.fast_writer, "        generate ugly Ksi code");
    ("-pretty", Clear Ksi.fast_writer, "      generate pretty Ksi code");
    ("-L", String (fun s -> link_search_path := 
                        (subst_plus s) :: !link_search_path),
        "DIR        add directory to library search path");
    ("-I", String (fun s -> iface_search_path := 
                        (subst_plus s) :: !iface_search_path),
        "DIR        add directory to interface search path");
    ("-link-aux", String (fun s -> link_aux := s), "NAME    set name \
                                of auxiliary ksi file generated during link");
    ("-link-init", String (fun s -> link_init := s), "NAME   set name of \
                                generated initialization function");
    ("-link-fini", String (fun s -> link_fini := s), "NAME   set name of \
                                generated finalization function");
    ("-link-main", String (fun s -> link_main := s), "NAME   set name of \
                           generated toplevel function (`main' by default)");
    ("-just-make-aux", Set just_make_aux, "   just make auxiliary linking \
                                              file and stop");
    ("-only-needed", String add_ungreedy_input_file, 
                     "LIBRARY  add only needed files from named library");
    ("-lib", String add_ungreedy_input_file, 
                     "LIBRARY  same as -only-needed");
    ("-greedy", String add_greedy_input_file, 
                     "LIBRARY  add all files from named library");
    ("-ldopt", String (fun s -> ldopts := s ^ " " ^ !ldopts),
                     " OPTION   pass option to C linker");
  ] in

  Arg.parse opt add_greedy_input_file help_str;
  ksi_opts := List.rev !ksi_opts;

  Gontmodule.set_iface_search_path (("." :: !iface_search_path) @ 
                                    [!prefix ^ "/stdlib/std.gia";
                                     !prefix ^ "/corelib/core.gia"]);

  let input = List.rev !input_files in
  let rec dumper f res = function
    | x :: xs -> 
      if f x != 0 then dumper f 1 xs
      else dumper f res xs
    | [] -> exit res in
  begin
    if input = [] then
      begin
        err "so what?";
        err ("run " ^ prog ^ " -help to get some info");
        exit 1
      end
    else if !dump_flag then
      dumper Gont.dump_module 0 (List.map snd input)
    else if !ciface_flag then
      dumper Ciface.dump_header 0 (List.map snd input)
    else
      match (!gia_mode, !ga_mode) with
      | (true, true) ->
          err "-gia and -ga are mutally exclusive";
          exit 1
      | (true, false) 
      | (false, true) ->
         begin
           match (List.map (fun (_, n) -> n) input) with
           | [] | [_] ->
               err "-ga and -gia flags need at least two arguments";
               err ("usage: " ^ prog ^ " -gia libname.gia iface1.gio...");
               err ("   or: " ^ prog ^ " -ga libname.ga file1.go...");
               exit 1
           | target::files ->
              try
                if !gia_mode then
                  Gontmodule.make_archive target files
                else
                  Gontlink.make_archive target !ldopts files
              with Gontutil.Error s -> begin
                err ("make_archive: " ^ s);
                rm target;
                exit 1
              end
         end
      | (false, false) ->
         try
           let out = List.map compile_file input in
           if !compile_only || !generate_ksi_only then
             ()
           else
             link out
         with Exit -> exit 1
  end;
  
  exit 0
;;

main ()
;;
